"""WAeUP data center.

The waeup data center cares for management of upload data and provides
tools for importing/exporting CSV data.
"""
import os
import shutil
import struct
import sys
import grok
from datetime import datetime
from zope.component import getUtility, getMultiAdapter
from zope.interface import Interface
from waeup.csvfile import getCSVFile
from waeup.interfaces import IWAeUPObject, IWAeUPCSVImporter, ICSVDataReceivers
from waeup.utils.helpers import copyFileSystemTree
from waeup.viewlets import Index, MainArea, LeftSidebar

class DataCenter(grok.Container):
    """A data center contains CSV files.
    """
    storage = os.path.join(os.path.dirname(__file__), 'files')

    def getReceivers(self):
        receivers = []
        curr_obj = getattr(self, '__parent__', None)
        while curr_obj is not None:
            if ICSVDataReceivers.providedBy(curr_obj):
                receivers = self.getCSVDataReceivers(curr_obj)
                break
            curr_obj = getattr(curr_obj, '__parent__', None)
        return receivers

    def getReceiverIds(self):
        """Get a dict of available receivers.

        The keys of the result are the receiver ids.
        """
        receivers = self.getReceivers()
        return dict([(self.getReceiverId(x), x) for x in receivers])
    
    def getImporters(self):
        """Get a list of all importers available.

        The search for available importers is done in two steps:

        1) Look for a utility providing ICSVDataReceiver,

        2) For every item of that utility: try to get an adapter
           providing IWAeUPCSVImporter.
        """
        result = []
        receivers = self.getReceivers()
        files = self.getFiles()
        for receiver in receivers:
            for file in files:
                wrapped_file = getCSVFile(file.context)
                try:
                    importer = getMultiAdapter((wrapped_file, receiver),
                                           IWAeUPCSVImporter)
                    result.append(importer)
                except:
                    # No multi-adapter for this combination available...
                    pass
        return result
    
    def getFiles(self):
        """Get a list of files stored in `storage`.
        """
        result = []
        if not os.path.exists(self.storage):
            return result
        for filename in os.listdir(self.storage):
            fullpath = os.path.join(self.storage, filename)
            if not os.path.isfile(fullpath):
                continue
            result.append(DataCenterFile(fullpath))
        return result

    def setStoragePath(self, path, move=False, overwrite=False):
        """Set the path where to store files.
        """
        path = os.path.abspath(path)
        not_copied = []
        if not os.path.exists(path):
            raise ValueError('The path given does not exist: %s' % path)
        if move is True:

            not_copied = copyFileSystemTree(self.storage, path,
                                            overwrite=overwrite)
        
        self.storage = path
        return not_copied

    def getCSVDataReceivers(self, obj):
        """Get a list of attributes, that can receive CSV data.
        """
        result = []
        for attr_name in dir(obj):
            if attr_name.startswith('_'):
                continue
            try:
                attr = getattr(obj, attr_name)
                # This might fail...
                #IWAeUPCSVImporter(attr)
                result.append(attr)
            except:
                pass
        return result

    def getPossibleImports(self):
        """Get list of possible imports.

        A single import is defined as a tuple
        
          ( <file-descr>, <importers> )
        
        where ``<file-descr>`` is an `IDataCenterFile` object and
        ``<importers>`` is a list of `IWAeUPCSVImporter` objects.
        """
        result = []
        importers = self.getImporters()
        for filedescr in self.getFiles():
            possible_importers = []
            for importer in importers:
                if importer.csvfile.path != filedescr.context:
                    continue
                importer_context = (
                    importer, self.getReceiverId(importer.receiver))
                possible_importers.append(importer_context)
            if len(possible_importers) == 0:
                continue
            result.append((filedescr, possible_importers))
        return result

    def getReceiverId(self, obj):
        """Get a unique id for an object.

        If the object is stored in ZODB it should contain a `_p_oid`
        attribute, which is guaranteed to be unique over the ZODB.

        If the object has no such attribute, then it will be held in
        memory only and will vanish as soon as the request is over. In
        this case we can use the memory address of it.

        Both, the id() result and the ZODB oid, are actually integers
        which we return as strings. To make the difference chrystal
        clear, we prepend ``z`` to ids taken from ZODB oids.
        """
        if not hasattr(obj, '_p_oid'):
            return str(id(obj))
        oid = getattr(obj, '_p_oid')
        if oid is None:
            # The persistent object is not stored in the ZODB
            return str(id(obj))
        return 'z%s' % struct.unpack('>Q', oid)[0]

    def doImport(self, path, receiverid, clear=None):
        receivers = self.getReceiverIds()
        if receiverid not in receivers.keys():
            raise ValueError('The requested data receiver cannot be found.')
        receiver = receivers[receiverid]
        filewrapper = getCSVFile(path)
        if filewrapper is None:
            raise ValueError('Format of CSV file not supported.')
        importer = getMultiAdapter((filewrapper, receiver), IWAeUPCSVImporter)
        if clear is not None:
            importer.doImport(clear_old_data=clear)
        else:
            importer.doImport()
        return

        
class IDataCenterFile(Interface):
    """A data center file.
    """
    pass


class DataCenterFile(object):
    """A description of a file stored in data center.
    """
    grok.implements(IDataCenterFile)
    
    def __init__(self, context):
        self.context = context
        self.name = os.path.basename(self.context)
        self.size = self.getSize()
        self.uploaddate = self.getDate()

    def getDate(self):
        """Get a human readable datetime representation.
        """
        date = datetime.fromtimestamp(os.path.getctime(self.context))
        return date.strftime('%c')

    def getSize(self):
        """Get a human readable file size.
        """
        bytesize = os.path.getsize(self.context)
        size = "%s bytes" % bytesize
        units = ['kb', 'MB', 'GB']
        for power, unit in reversed(list(enumerate(units))):
            power += 1
            if bytesize >= 1024 ** power:
                size = "%.2f %s" % (bytesize/(1024.0**power), unit)
                break
        return size


class Content(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(Index)

class Upload(grok.View):
    grok.context(DataCenter)
    grok.name('upload')
    grok.template('master')

    def update(self, uploadfile=None, CANCEL=None, SUBMIT=None):
        if CANCEL is not None:
            self.redirect(self.url(self.context))
            return
        if not uploadfile:
            return
        try:
            filename = uploadfile.filename
            target = os.path.join(self.context.storage, filename)
            open(target, 'wb').write(uploadfile.read())
        except IOError:
            self.flash('Error while uploading file. Please retry.')
            self.flash('I/O error: %s' % sys.exc_info()[1])
            return
        self.redirect(self.url(self.context))
    
class UploadMain(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(Upload)
    

class Settings(grok.View):
    grok.context(DataCenter)
    grok.name('settings')
    grok.template('master')

    def update(self, newpath=None, move=False, overwrite=False,
               save=None, cancel=None):
        if move:
            move = True
        if overwrite:
            overwrite = True
        if newpath is None:
            return
        if cancel is not None:
            self.redirect(self.url(self.context))
            return
        try:
            not_copied = self.context.setStoragePath(newpath, move=move)
            for name in not_copied:
                self.flash('File already existed (not copied): %s' % name)
        except ValueError:
            self.flash('Given storage path cannot be used.')
            self.flash('Error: %s' %sys.exc_info()[1])
            return
        if newpath:
            self.flash('New storage path succefully set.')
            self.redirect(self.url(self.context))
        return

class SettingsMain(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(Settings)

class SettingsLink(grok.Viewlet):
    grok.viewletmanager(LeftSidebar)
    grok.context(DataCenter)
    grok.view(Index)
    grok.order(2)
    grok.require('waeup.manageUniversity')

    def render(self):
        return u'<div class="portlet"><a href="@@settings">Settings</a></div>'

class ImportCSV(grok.View):
    """A view for importing things.

    """
    grok.require('waeup.manageUniversity')
    grok.template('master')
    grok.context(DataCenter)

    def update(self, csvfile=None, clear=None, overwrite=None,
               receiverid=None, CANCEL=None, SUBMIT=None):
        if CANCEL is not None:
            self.redirect(self.url(self.context))
            return
        if not csvfile:
            return
        if not SUBMIT:
            return
        try:
            self.context.doImport(csvfile, receiverid, clear=clear)
        except ValueError:
            self.flash('Could not import: %s' % os.path.basename(csvfile))
            return
        except:
            self.flash('Import failed: %s' % sys.exc_info()[1])
            return
        self.flash('Successfully imported: %s' % os.path.basename(csvfile))
        self.redirect(self.url(self.context))
        return

class Import(object):
    """Helper class to aggregate imports and their data.
    """
    def __init__(self, filedescr, importers):
        self.file = filedescr
        self.importers = []
        for importer, receiverid in importers:
            importer.receiverid = receiverid
            self.importers.append(importer)
        
class ImportCSVMain(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(ImportCSV)
    grok.require('waeup.manageUniversity')

    def getImports(self):
        result = []
        imports = self.context.getPossibleImports()
        result = [Import(x, y) for x, y in imports]
        return result
    
    def update(self, *args, **kw):
        return super(ImportCSVMain, self).update(*args, **kw)
        
